//+------------------------------------------------------------------+
//|                                                PerceptronXOR.mq5 |
//|                                     Copyright 2020, Lethan Corp. |
//|                           https://www.mql5.com/pt/users/14134597 |
//+------------------------------------------------------------------+
#property copyright "Copyright 2020, Lethan Corp."
#property link      "https://www.mql5.com/pt/users/14134597"
#property version   "1.00"
//#property script_show_inputs

#include "Libraries\Util.mqh"
#include <Arrays\ArrayDouble.mqh>
#include <Graphics\Graphic.mqh>

#define nINPUT 4

sinput int look_back = 300;
CGraphic myGraphic[1];
int handleMA;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---
   random.seed(42);
   double dataset[][nINPUT];
   ArrayResize(dataset, look_back*nINPUT);

   double close[];
   double ma[];
   CArrayDouble *lab = new CArrayDouble();
   CArrayDouble *ret = new CArrayDouble();

   int width = (int)ChartGetInteger(0, CHART_WIDTH_IN_PIXELS);
   int height = (int)ChartGetInteger(0, CHART_HEIGHT_IN_PIXELS);
   myGraphic[0].Create(0,"Curve",0,20,20, width, height);

   handleMA = iMA(NULL,PERIOD_H1,20,0,MODE_SMA,PRICE_CLOSE);
   CopyClose(Symbol(),PERIOD_H1,0,look_back*nINPUT+nINPUT-1,close);
   CopyBuffer(handleMA, 0, 0, look_back*nINPUT+nINPUT-1, ma);
   lab = add_labels(close);

   ArrayReverse(close);
   ArrayReverse(ma);

   int idx=0;
   for(int row=0; row<look_back; row++)
     {
      for(int i=0; i<nINPUT; i++)
        {
         if(lab[idx]==DBL_MIN || lab[idx]==DBL_MAX)
           {
            ArrayResize(dataset, ArrayRange(dataset, 0)-1);
            ArrayResize(close, ArrayRange(close, 0)-1);
            continue;
           }
         //dataset[idx][0]= close[idx];
         dataset[idx][0]= close[idx] - ma[idx];
         dataset[idx][1]= close[idx+1] - ma[idx+1];
         dataset[idx][2]= close[idx+2] - ma[idx+2];
         dataset[idx][3]= lab[idx];
         idx+=1;
        }
     }
   ArrayReverse(dataset);
   ArrayReverse(close);

   double tester[][2];
   for(int i=0; i<ArrayRange(dataset, 0); i++)
     {
      ArrayResize(tester, i+1);
      tester[i][0]=close[i];
      tester[i][1]=dataset[i][3];
     }

   ret = Tester(tester);
   double plot[];
   ArrayResize(plot,ret.Total());
   for(int i=0; i<ret.Total(); i++)
      plot[i]=ret[i];

   myGraphic[0].CurveRemoveByName("Curve");
   myGraphic[0].CurveAdd(plot, clrBlueViolet, CURVE_LINES,"Curve");
   myGraphic[0].CurvePlotAll();
   myGraphic[0].Update();

   Sleep(5000);

   double weights[];
   train_weights(weights, dataset, 0.1, 100);
   double validate[][2];
   for(int row=0; row<ArrayRange(dataset, 0); row++)
     {
      double predict = predict(dataset, weights, row);
      printf("Expected=%.1f, Predicted=%.1f", dataset[row][nINPUT-1], predict);

      ArrayResize(validate, row+1);
      validate[row][0]=close[row];
      validate[row][1]=predict;
     }

   ArrayFree(plot);
   ret = Tester(validate);
   ArrayResize(plot,ret.Total());
   for(int i=0; i<ret.Total(); i++)
      plot[i]=ret[i];

   myGraphic[0].CurveRemoveByName("Curve");
   myGraphic[0].CurveAdd(plot, clrBlueViolet, CURVE_LINES,"Curve");
   myGraphic[0].CurvePlotAll();
   myGraphic[0].Update();

   Sleep(5000);

   myGraphic[0].CurveRemoveByName("Curve");
   myGraphic[0].Destroy();

  }
//+------------------------------------------------------------------+
// Make a prediction with weights
template <typename Array>
double predict(const Array &X[][nINPUT], const Array &weights[], const int row=0)
  {
   double z = weights[0];
   for(int i=0; i<ArrayRange(X, 1)-1; i++)
     {
      z+=weights[i+1]*X[row][i];
     }
   return activation(z);
  }
//+------------------------------------------------------------------+
//|                Transfer neuron activation                        |
//+------------------------------------------------------------------+
double activation(const double activation) //#
  {
   return activation>=0.0?1.0:0.0;
  }
//+------------------------------------------------------------------+
//|  Estimate Perceptron weights using stochastic gradient descent   |
//+------------------------------------------------------------------+
template <typename Array>
void train_weights(Array &weights[], const Array &X[][nINPUT], double l_rate=0.1, int n_epoch=5)
  {
   ArrayResize(weights, ArrayRange(X, 1));
   for(int i=0; i<ArrayRange(X, 1); i++)
     {
      weights[i]=random.random();
     }

   for(int epoch=0; epoch<n_epoch; epoch++)
     {
      double sum_error = 0.0;
      for(int row=0; row<ArrayRange(X, 0); row++)
        {
         double y = predict(X, weights, row);
         double error = X[row][nINPUT-1] - y;
         sum_error += pow(error, 2);
         weights[0] = weights[0] + l_rate * error;

         for(int i=0; i<ArrayRange(X, 1)-1; i++)
           {
            weights[i+1] = weights[i+1] + l_rate * error * X[row][i];
           }
        }
      printf(">epoch=%d, lrate=%.3f, error=%.3f",epoch, l_rate, sum_error);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CArrayDouble *add_labels(double &dataset[], int min=1, int max=25)
  {
   CArrayDouble *labels=new CArrayDouble();
   for(int i=0; i<ArrayRange(dataset, 0)-max; i++)
     {
      int rand = random.randint(min, max);
      if(dataset[i] >= (dataset[i + rand]))
        {
         labels.Add(1.0);
        }
      else
        {
         if(dataset[i] <= (dataset[i + rand]))
           {
            labels.Add(0.0);
           }
         else
           {
            labels.Add(0.0);
           }
        }
     }
   return(GetPointer(labels));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CArrayDouble *Tester(double &dataset[][2], double markup = 0.0)
  {
   int last_deal = INT_MAX;
   double last_price = 0.0;
   CArrayDouble *report=new CArrayDouble();
   report.Add(0.0);
   for(int i=0; i<ArrayRange(dataset, 0); i++)
     {
      double pred = dataset[i][1];
      if(last_deal > 1)
        {
         last_price = dataset[i][0];
         last_deal = pred<=0.5?0:1;
         continue;
        }

      if(last_deal == 0 && pred > 0.5)
        {
         last_deal = 1;
         report.Add(report[report.Total()-1] - markup + (dataset[i][0] - last_price));
         last_price = dataset[i][0];
         continue;
        }

      if(last_deal == 1 && pred <=0.5)
        {
         last_deal = 0;
         report.Add(report[report.Total()-1] - markup + (last_price - dataset[i][0]));
         last_price = dataset[i][0];
         continue;
        }
      else
        {
         report.Add(report[report.Total()-1]);
        }
     }
   return(GetPointer(report));
  }